/*
 * (C) 2009 The University of Chicago
 *
 * See COPYRIGHT in top-level directory.
 */

#ifndef __TRITON_BASE_H__
#define __TRITON_BASE_H__

/**
 * triton-base.h includes important system headers and defines appropriate
 * macros used throughout.
 */

#ifdef __STRICT_ANSI__
/* strict ANSI C doesn't have 'inline' functions, so we use __inline__ instead */
#define inline __inline__ __attribute__((flatten))
#endif

/* simplified version of the unused attribute so that gcc doesn't give warning messages */
#ifndef __unused__
#define __unused__ __attribute__((unused))
#endif

/* older systems have malloc in separate header */
#include <stdlib.h>

/* basic system headers */
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>

/* macros for portable printing of 64 bit values */
#if 0
/* this version relies on configure-time results in config.h */
#ifndef llu
#if SIZEOF_LONG_INT == 4 
#  define llu(x) (x)
#  define lld(x) (x)
#  define SCANF_lld lld
#elif SIZEOF_LONG_INT == 8
#  define llu(x) (unsigned long long)(x)
#  define lld(x) (long long)(x)
#  define SCANF_lld ld
#else
#  error Unexpected sizeof(long int)
#endif
#endif
#else

/* this version relies on less-portable system definitions */
#ifndef llu
#if __WORDSIZE == 64
#  define llu(x) (unsigned long long)(x)
#  define lld(x) (long long)(x)
#  define SCANF_lld "%ld"
#else
#  define llu(x) (x)
#  define lld(x) (x)
#  define SCANF_lld "%lld"
#endif
#endif

#endif

/* set of macros that allow us to count the number
 * of arguments passed into a CPP macro that takes
 * a variable number of arguments.  This, in turn,
 * allows us to use variable arguments in C in a way
 * that allows us to not need to pass in a specific
 * count field.
 *
 * Some examples:
 *
 * COUNT_ARGS() -> 0
 * COUNT_ARGS(A) -> 1 
 * COUNT_ARGS(A,B) -> 2 
 * COUNT_ARGS(A,B,C) -> 3 
 * 
 * Can be as many as 15 arguments.  For examples of
 * how this is used, look at the triton_init() function
 * and associated macro.
 *
 * These macros assume an important feature of GCC that is not part of C99, namely,
 * by adding the ## token in front of __VA_ARGS__, if the arguments list is empty,
 * a preceding comma is stripped from the output.  See: http://gcc.gnu.org/onlinedocs/cpp/Variadic-Macros.html
 * for more details.
 *
 * Taken from a post by Laurent Deniau on comp.std.c (Jan, 2006).
 * and to allow for zero arguments:  https://gustedt.wordpress.com/2010/06/08/detect-empty-macro-arguments/
 */
#define _TRIGGER_PARENS_(...) X, X, X, X, X, X, X, X, X, X, X, X, X, X, X, X
#define _ARG16(_0, _1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11, _12, _13, _14, _15, _16, ...) _16
#define COUNT_ARGS_(...) _ARG16(__VA_ARGS__, 0, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1)
#define COUNT_ARGS(...) COUNT_ARGS_(_TRIGGER_PARENS_ __VA_ARGS__ ())

#define _TRIGGER_PARENS_(...) X, X, X, X, X, X, X, X, X, X, X, X, X, X, X, X
#define _FORMAT_ARG16(_0, _1, _2, _3, _4, _5, _6, _7, _8, _9, _10, _11, _12, _13, _14, _15, _16, ...) SREPEAT("%s", _16), _16
#define FORMAT_ARGS_(...) _FORMAT_ARG16(__VA_ARGS__, 0, 15, 14, 13, 12, 11, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1)
#define FORMAT_ARGS(...) FORMAT_ARGS_(_TRIGGER_PARENS_ __VA_ARGS__ ())

/* repeats a format string a number of times:
 * 
 * SREPEAT("%s", 4) -> "%s%s%s%s"
 */
#define SREPEAT0(x) " "
#define SREPEAT1(x) x
#define SREPEAT2(x) x x
#define SREPEAT3(x) x x x
#define SREPEAT4(x) x x x x
#define SREPEAT5(x) x x x x x
#define SREPEAT6(x) x x x x x x
#define SREPEAT7(x) x x x x x x x
#define SREPEAT8(x) x x x x x x x x
#define SREPEAT9(x) x x x x x x x x x
#define SREPEAT10(x) x x x x x x x x x x
#define SREPEAT11(x) x x x x x x x x x x x
#define SREPEAT12(x) x x x x x x x x x x x x
#define SREPEAT13(x) x x x x x x x x x x x x x
#define SREPEAT14(x) x x x x x x x x x x x x x x
#define SREPEAT15(x) x x x x x x x x x x x x x x x

#define SREPEAT_(x,i) SREPEAT##i(x)
#define SREPEAT(x,i) SREPEAT_(x,i)

#endif /* __TRITON_BASE_H__ */

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 *
 * vim: ft=c ts=8 sts=4 sw=4 expandtab
 */
