/*
 * (C) 2009 The University of Chicago
 *
 * See COPYRIGHT in top-level directory.
 */

#ifndef TRITON_HASH_H
#define TRITON_HASH_H


#include "triton-list.h"
#include "jenkins-hash.h"

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#define triton_hash_link                 triton_list_link

/* __typename__ * triton_hash_get_entry(struct triton_hash_link *link, __typename__, __link_field_name__); */
#define triton_hash_get_entry            triton_list_get_entry

struct triton_hash_bucket
{
    triton_list_t entries;
    void *external;
};

struct triton_hash_table
{
    struct triton_hash_bucket *array;
    int table_size;
    int (*compare) (const void *key, struct triton_hash_link * link);
    int (*hash) (const void *key, int table_size);
    int (*bucket_init)(struct triton_hash_bucket *bucket);
    void (*bucket_destroy)(struct triton_hash_bucket *bucket);
};

static inline int triton_hash_bucket_init_default(struct triton_hash_bucket *bucket)
{
    triton_list_init(&bucket->entries);
    bucket->external = NULL;
    return 0;
};

/* triton_hash_init()
 *
 * creates a new hash table with the specified table size.  The
 * hash function and compare function must be provided.
 * table_size should be a good prime number.
 *
 * returns pointer to table on success, NULL on failure
 */
static inline struct triton_hash_table *triton_hash_init_general(
    int (*compare) (const void *key,
		    struct triton_hash_link * link),
    int (*hash) (const void *key,
		 int table_size),
    int table_size,
    int (*bucket_init)(struct triton_hash_bucket *bucket),
    void (*bucket_destroy)(struct triton_hash_bucket *bucket))
{
    int i = 0;
    struct triton_hash_table *new_table = NULL;

    /* create struct to contain table information */
    new_table = (struct triton_hash_table *)malloc(sizeof(struct triton_hash_table));
    if (!new_table)
    {
	return (NULL);
    }

    /* fill in info */
    new_table->compare = compare;
    new_table->hash = hash;
    new_table->table_size = table_size;
    new_table->bucket_init = bucket_init;
    new_table->bucket_destroy = bucket_destroy;

    /* create array for actual table */
    new_table->array = (struct triton_hash_bucket *)
	malloc(sizeof(*new_table->array) * table_size);
    if (!new_table->array)
    {
	free(new_table);
	return (NULL);
    }

    if(bucket_init)
    {
        /* initialize a doubly linked at each hash table index */
        for (i = 0; i < table_size; i++)
        {
            bucket_init(&new_table->array[i]);
        }
    }

    return (new_table);
}

static inline struct triton_hash_table *triton_hash_init(
    int (*compare) (const void *key,
		    struct triton_hash_link * link),
    int (*hash) (const void *key,
		 int table_size),
    int table_size)
{
    return triton_hash_init_general(compare, hash, table_size, triton_hash_bucket_init_default, NULL);
}


/* triton_hash_finalize()
 *
 * frees any resources created by the hash table
 *
 * no return value
 */
static inline void triton_hash_finalize(
    struct triton_hash_table *old_table)
{
    int i;

    if(old_table->bucket_destroy)
    {
        /* initialize a doubly linked at each hash table index */
        for (i = 0; i < old_table->table_size; i++)
        {
            old_table->bucket_destroy(&old_table->array[i]);
        }
    }

    free(old_table->array);
    free(old_table);
    return;
}

static inline struct triton_hash_bucket *triton_hash_to_bucket(
    struct triton_hash_table *table,
    const void *key)
{
    int hindex = 0;

    assert(table);
    assert(key);

    /* find the hash value */
    hindex = table->hash(key, table->table_size);
    assert(hindex >= 0 && hindex < table->table_size);
    return &table->array[hindex];
}

static inline struct triton_hash_link *triton_hash_bucket_to_entry(
    struct triton_hash_table *table,
    struct triton_hash_bucket *bucket,
    const void *key)
{
    struct triton_hash_link *tmp_link, *safe;

    assert(bucket);

    /* linear search at hindex to find match */
    triton_list_for_each(tmp_link, safe, &bucket->entries)
    {
	if (table->compare(key, tmp_link))
	{
	    return (tmp_link);
	}
    }
    return (NULL);
}

static inline void triton_hash_add_to_bucket(
    struct triton_hash_table *table,
    struct triton_hash_bucket *bucket,
    struct triton_hash_link *llink)
{
    assert(bucket != NULL);
    triton_list_add_back(llink, &bucket->entries);
}

/* triton_hash_add()
 *
 * adds a new link onto the hash table, hashes based on given key
 *
 * no return value
 */
static inline void triton_hash_add(
    struct triton_hash_table *table,
    const void *key,
    struct triton_hash_link *llink)
{
    struct triton_hash_bucket *b;
    b = triton_hash_to_bucket(table, key);
    triton_hash_add_to_bucket(table, b, llink);
}

/* triton_hash_search()
 *
 * searches for a link in the hash table
 * that matches the given key
 *
 * returns pointer to link on success, NULL on failure (or item not found)
 */
static inline struct triton_hash_link *triton_hash_search(
    struct triton_hash_table *table,
    const void *key)
{
    return triton_hash_bucket_to_entry(table, triton_hash_to_bucket(table, key), key);
}

static inline struct triton_hash_link *triton_hash_remove_entry_at_bucket(
    struct triton_hash_table *table,
    struct triton_hash_bucket *bucket,
    const void *key)
{
    struct triton_hash_link *tmp_link, *safe;

    if(triton_list_empty(&bucket->entries))
    {
        return NULL;
    }

    /* linear search at hindex to find match */
    triton_list_for_each(tmp_link, safe, &bucket->entries)
    {
	if (table->compare(key, tmp_link))
	{
	    triton_list_del(tmp_link);
	    return (tmp_link);
	}
    }
    return (NULL);
}

/* triton_hash_search_and_remove()
 *
 * searches for and removes a link in the hash table
 * that matches the given key
 *
 * returns pointer to link on success, NULL on failure (or item
 * not found).  On success, link is removed from hashtable.
 */
static inline struct triton_hash_link *triton_hash_search_and_remove(
    struct triton_hash_table *table,
    const void *key)
{
    return triton_hash_remove_entry_at_bucket(table, triton_hash_to_bucket(table, key), key);
}

#define triton_hash_destroy_entries(                                                            \
    _oldtable, _entry_type, _link, _destructor)                                                 \
    do                                                                                          \
    {                                                                                           \
        int __dfi = 0;                                                                          \
        struct triton_list_link *__dfentry;                                                     \
        struct triton_list_link *__dftmpe;                                                      \
        for(__dfi = 0; __dfi < _oldtable->table_size; ++__dfi)                                  \
        {                                                                                       \
            triton_list_for_each(__dfentry, __dftmpe, &(_oldtable->array[__dfi].entries))       \
            {                                                                                   \
                triton_list_del(__dfentry);                                                     \
                _destructor(triton_hash_get_entry(__dfentry, _entry_type, _link));              \
            }                                                                                   \
        }                                                                                       \
    } while(0)
 
#define triton_hash_destroy_and_finalize(                                        \
    _oldtable, _entry_type, _link, _destructor)                                  \
    do                                                                           \
    {                                                                            \
        triton_hash_destroy_entries(_oldtable, _entry_type, _link, _destructor); \
        triton_hash_finalize(_oldtable);                                         \
                                                                                 \
    } while(0)

#define triton_hash_for_each_bucket(_table, _bucket)              \
        for((_bucket) = (_table)->array;                          \
            ((_bucket)-((_table)->array))                         \
            < (_table)->table_size;                               \
            ++(_bucket))

#define triton_hash_for_each_entry_in_bucket(__bucket, __entry, __safe, __type, __member) \
    triton_list_for_each_entry(__entry, __safe, &(__bucket)->entries, __type, __member)

#define triton_hash_map(__table, __func, __up, __type, __member)                            \
    do {                                                                                    \
        struct triton_hash_bucket *__map_bucket;                                            \
        triton_hash_for_each_bucket(__table, __map_bucket)                                  \
        {                                                                                   \
            __type *e, *s;                                                                  \
            triton_hash_for_each_entry_in_bucket(__map_bucket, e, s, __type, __member)      \
            {                                                                               \
                __func(__up, e);                                                            \
            }                                                                               \
        }                                                                                   \
    } while(0)

/* if your key is already has good hashing properties, just use it as is */
static inline int triton_hash_no_hash(const void *k, int table_size)
{
    const uint32_t *pc = k;
    return (*pc) & (table_size - 1);
}

static inline int triton_hash_32bit_hash(const void *k, int table_size)
{
    uint32_t pc = 0, pb = 0;
    bj_hashlittle2(k, 4, &pc, &pb);
    return pc & (table_size - 1);
}

static inline int triton_hash_64bit_hash(const void *k, int table_size)
{
    uint32_t pc = 0, pb = 0;
    bj_hashlittle2(k, 8, &pc, &pb);
    return (int) (pc & (table_size - 1));
}

static inline int triton_hash_128bit_hash(const void *k, int table_size)
{
    uint32_t pc = 0, pb = 0;
    bj_hashlittle2(k, 16, &pc, &pb);
    return (int) (pc & (table_size - 1));
}

static inline int triton_hash_strhash(const void *k, int table_size)
{
    uint32_t pc = 0, pb = 0;
    bj_hashlittle2(k, strlen((const char *)k), &pc, &pb);
    return pc & (table_size - 1);
}

// Hashes const char * 
// Note: cannot handle embedded 0 (C string limitation)
static inline int triton_hash_string_hash(const void *k, int table_size)
{
    uint32_t pc = 0, pb = 0;
    const char *s;
    int result = -1;
    s = (const char *)k;
    assert(s);
    bj_hashlittle2(s, strlen(s), &pc, &pb);
    result = pc & (table_size - 1);
    return result;
}

static inline int triton_hash_ptr_hash(const void *k, int table_size)
{
    uint32_t pc = 0, pb = 0;
    bj_hashlittle2(&k, sizeof(k), &pc, &pb);
    return pc & (table_size - 1);
}

#endif /* TRITON_HASH_H */

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 *
 * vim: ft=c ts=8 sts=4 sw=4 expandtab
 */
