/*
 * (C) 2009 The University of Chicago
 *
 * See COPYRIGHT in top-level directory.
 */

#ifndef __TRITON_LIST_H__
#define __TRITON_LIST_H__

#include <stdint.h>
#include <assert.h>
#include <stddef.h>

typedef struct triton_list triton_list_t;

typedef struct triton_list_link
{
    triton_list_t *inlist;
    struct triton_list_link *next;
    struct triton_list_link *prev;
} triton_list_link_t;

struct triton_list
{
    struct triton_list_link entries;
    uint64_t count;
};

#define triton_list_tolink(__list) (&(__list)->entries)

#define TRITON_LIST_NULL { { NULL, NULL, NULL }, 0 }

#define TRITON_LIST_STATIC_INITIALIZER(name) { { &name, &(name.entries), &(name.entries) }, 0 }

#define TRITON_LIST_DEFINE(name) \
    triton_list_t name = TRITON_LIST_STATIC_INITIALIZER(name)

static inline void triton_list_link_clear(triton_list_link_t *llink)
{
    ((struct triton_list_link *)llink)->inlist = NULL;
    ((struct triton_list_link *)llink)->next = NULL;
    ((struct triton_list_link *)llink)->prev = NULL;
}

static inline void triton_list_init(triton_list_t *list)
{
    list->entries.next = (struct triton_list_link *)&list->entries;
    list->entries.prev = (struct triton_list_link *)&list->entries;
    list->entries.inlist = list;
    list->count = 0;
}

/*
 * Insert a new entry between two known consecutive entries. 
 *
 * This is only for internal triton_list manipulation where we know
 * the prev/next entries already!
 */
static inline void __triton_list_add(struct triton_list_link * new,
                                 struct triton_list_link * prev,
                                 struct triton_list_link * next)
{
    assert(new->inlist == NULL);
    assert(new->next == NULL && new->prev == NULL);
    next->prev = new;
    new->next = next;
    new->prev = prev;
    prev->next = new;
}

static inline void triton_list_add_front(
    triton_list_link_t *new, triton_list_t *list)
{
    __triton_list_add(new, &list->entries, list->entries.next);
    new->inlist = list;
    list->count++;
}

/**
 * Return NULL if this entry does not belong to a list,
 * or a pointer to the list if it does.
 */
static inline triton_list_t * triton_list_return_list (triton_list_link_t * n)
{
   return n->inlist;
}

static inline void triton_list_add_back(
    triton_list_link_t *new, triton_list_t *list)
{
    __triton_list_add(new, list->entries.prev, &list->entries);
    new->inlist = list;
    list->count++;
}

static inline void __triton_list_del(struct triton_list_link * prev,
                                 struct triton_list_link * next)
{
    next->prev = prev;
    prev->next = next;
}

static inline void triton_list_del(struct triton_list_link *entry)
{
    assert(entry->prev != NULL && entry->next != NULL);
    __triton_list_del(entry->prev, entry->next);
    entry->prev = NULL; entry->next = NULL;
    entry->inlist->count--;
    entry->inlist = NULL;
}

/**
 * Returns non-zero if the given link is in a linked list.
 * Returns 0 if the link is not linked in a list.
 */
static inline int triton_list_is_linked (struct triton_list_link * e)
{
   return e->inlist != NULL;
}

static inline int triton_list_empty(triton_list_t *list)
{
    return list->count == 0;
}

/**
 * Clear the list
 */
static inline void triton_list_clear (triton_list_t * list)
{
   triton_list_init (list);
}

/**
 * Destroy the list.
 * List must be empty
 */
static inline void triton_list_destroy (triton_list_t * list)
{
   assert (triton_list_empty (list));
}

static inline uint64_t triton_list_count(triton_list_t *list)
{
    return list->count;
}

#define triton_list_get_entry(ptr, type, member) \
    ((type *)((char *)(ptr)-(unsigned long)((&((type *)0)->member))))

/**
 * triton_list_for_each - iterate over a list safe against removal of list entry
 * @pos:  the struct triton_list_link * to use as a loop counter.
 */
#define triton_list_for_each(__pos, __scratch, __list) \
    for (__pos = (triton_list_tolink(__list))->next, \
         __scratch = (triton_list_tolink(__list))->next->next; \
         __pos != (triton_list_tolink(__list)); \
         __pos = __scratch, __scratch = (__pos)->next)

/**
 * triton_list_for_each_entry_reverse - 
 *     iterate over list of given type safe against removal of list entry
 *     running from back to front of list
 * @pos:	the type * to use as a loop counter.
 * @head:	the head for your list.
 * @member:	the name of the list_struct within the struct.
 */
#define triton_list_for_each_entry_reverse(__pos, __scratch, __list, __type, __member) \
    for (__pos = triton_list_get_entry((triton_list_tolink(__list))->prev, __type, __member), \
         __scratch = triton_list_get_entry((triton_list_tolink(__list))->prev->prev, __type, __member); \
         (__pos) && (&(__pos)->__member != (triton_list_tolink(__list))); \
         __pos = __scratch, \
         __scratch = triton_list_get_entry((__pos)->__member.prev, __type, __member))


/**
 * triton_list_for_each_entry - 
 *     iterate over list of given type safe against removal of list entry
 * @pos:	the type * to use as a loop counter.
 * @head:	the head for your list.
 * @member:	the name of the list_struct within the struct.
 */
#define triton_list_for_each_entry(__pos, __scratch, __list, __type, __member) \
    for (__pos = triton_list_get_entry((triton_list_tolink(__list))->next, __type, __member), \
         __scratch = triton_list_get_entry((triton_list_tolink(__list))->next->next, __type, __member); \
         (__pos) && (&(__pos)->__member != (triton_list_tolink(__list))); \
         __pos = __scratch, \
         __scratch = triton_list_get_entry((__pos)->__member.next, __type, __member))


static inline void triton_list_moveto (triton_list_t * dest, triton_list_t *
      source, triton_list_link_t * sentinel)
{
   triton_list_init (dest);

   struct triton_list_link * mypos = source->entries.next;

   while (mypos != sentinel)
   {
      mypos->inlist = dest;
      mypos = mypos->next;
   }

   dest->entries.next = (source->entries.next == sentinel ?
                                &dest->entries
                              : source->entries.next);
   dest->entries.prev = (source->entries.prev == sentinel ?
                                &dest->entries
                              : source->entries.prev);

   dest->count = source->count;

   if (dest->count)
   {
      assert (dest->entries.next != &dest->entries);
      assert (dest->entries.prev != &dest->entries);
      dest->entries.next->prev = &dest->entries;
      dest->entries.prev->next = &dest->entries;
   }

   triton_list_init (source);
}

/**
 * Swap the contents of the given lists. Note: linear in time due to the
 * inlist link.
 */

static inline void triton_list_swap (triton_list_t * list1,
      triton_list_t * list2)
{
   struct triton_list tmp = *list1;

   triton_list_moveto (list1, list2, &list2->entries);
   triton_list_moveto (list2, &tmp, &list1->entries);
}

static inline int triton_list_exists(triton_list_t *list, struct triton_list_link *llink)
{
    struct triton_list_link *mypos, *safe;
    triton_list_for_each(mypos, safe, list)
    {
        if(mypos == llink)
        {
            return 1;
        }
    }
    return 0;
}

static inline struct triton_list_link * triton_list_find(
    triton_list_t *list,
    int (*compare)(struct triton_list_link *, void *),
    void *ptr)
{
    struct triton_list_link *pos, *safe;
    triton_list_for_each(pos, safe, list)
    {
        if(compare(pos, ptr))
        {
            return pos;
        }
    }
    return NULL;
}

static inline void triton_list_insert_after(
    triton_list_link_t *new,
    triton_list_link_t *after,
    triton_list_t *list)
{
    assert(new->inlist == NULL);
    assert(new->next == NULL && new->prev == NULL);
    assert(after->inlist == list);
    __triton_list_add(new, after, after->next);
    new->inlist = list;
    list->count++;
}

static inline void triton_list_insert_before(
    triton_list_link_t *new,
    triton_list_link_t *before,
    triton_list_t *list)
{
    assert(new->inlist == NULL);
    assert(new->next == NULL && new->prev == NULL);
    assert(before->inlist == list);
    __triton_list_add(new, before->prev, before);
    new->inlist = list;
    list->count++;
}


static inline struct triton_list_link * __triton_list_del_front(triton_list_t *list)
{
    struct triton_list_link *entry = list->entries.next;
    assert(entry->inlist == list);
    if(triton_list_empty(list)) return NULL;
    triton_list_del(list->entries.next);
    return entry;
}

static inline struct triton_list_link * triton_list_del_front(triton_list_t *list)
{
   return __triton_list_del_front (list);
}

static inline struct triton_list_link *__triton_list_peek_front(triton_list_t *list)
{
    if(triton_list_empty(list)) return NULL;
    return list->entries.next;
}

static inline struct triton_list_link * triton_list_peek_front (
      triton_list_t * list)
{
   return __triton_list_peek_front (list);
}

#define triton_queue_enqueue(__entry, __list) triton_list_add_back((__entry), (__list))
#define triton_queue_dequeue(__list) __triton_list_del_front(__list)
#define triton_queue_peek(__list) __triton_list_peek_front(__list)

#define triton_stack_push(__entry, __list) triton_list_add_front((__entry), (__list))
#define triton_stack_pop(__list) __triton_list_del_front(__list)
#define triton_stack_peek(__list) __triton_list_peek_front(__list)

#endif /* __TRITON_LIST_H__ */

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 *
 * vim: ft=c ts=8 sts=4 sw=4 expandtab
 */
