/*
 * (C) 2009 The University of Chicago
 *
 * See COPYRIGHT in top-level directory.
 */

#ifndef __TRITON_UINT128_H__
#define __TRITON_UINT128_H__

#include <stdio.h>
#include <assert.h>
#include <stdint.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <string.h>
#include <endian.h>

/* unsigned 128 bit integer type */
typedef struct { uint64_t l; uint64_t u; } uint128_t;

#define triton_uint128_setzero(t) (t).l = 0, (t).u = 0
#define triton_uint128_iszero(t) ((t).l == 0 && (t).u == 0)
#define triton_uint128_setmax(t) (t).l = UINT64_MAX, (t).u = UINT64_MAX
#define triton_uint128_ismax(t) ((t).l == UINT64_MAX && (t).u == UINT64_MAX)

#define triton_uint128_set(from, to) ((to).l = (from).l, (to).u = (from).u)

/* maximum true value of unsigned 128 bit integer is:
 * 340282366920938463463374607431768211455
 *
 * but we will actually print it as two 64 bit integers:
 * 18446744073709551615:18446744073709551615
 *
 * need 41 characters to print, 42 including terminating \0
 */
#define TRITON_UINT128_STRLEN 42

/* this version relies on less-portable system definitions */
#ifndef llu
#if __WORDSIZE == 64
#  define llu(x) (unsigned long long)(x)
#  define lld(x) (long long)(x)
#  define SCANF_lld ld
#else
#  define llu(x) (x)
#  define lld(x) (x)
#  define SCANF_lld lld
#endif
#endif

/* convert uint128 to string which is in ipv6 notation,
 * NULL on failure,
 * user must free buffer when done */
static inline char* triton_uint128_to_string(uint128_t big)
{
    char *buf = malloc(INET6_ADDRSTRLEN);
    const char *ptr;

    if (buf)
    {
        struct in6_addr ipv6_addr;
        uint64_t *upper = (uint64_t *) &ipv6_addr.s6_addr[0];
        uint64_t *lower = (uint64_t *) &ipv6_addr.s6_addr[8];
        *upper = htobe64(big.u);
        *lower = htobe64(big.l); 
        ptr = inet_ntop(AF_INET6, &ipv6_addr, buf, INET6_ADDRSTRLEN);
        if (!ptr)
        {
            free(buf); /* free buffer on failure */
            buf = NULL;
        }
    }

    return buf;
}

/* convert a string to a uint128_t, string is in ivp6 notation.
   returns max oid on failure */
static inline uint128_t triton_uint128_from_string (const char* str)
{
    struct in6_addr ipv6_addr;
    uint128_t big;
    int r;
    uint64_t *upper = (uint64_t *) &ipv6_addr.s6_addr[0];
    uint64_t *lower = (uint64_t *) &ipv6_addr.s6_addr[8];

    r = inet_pton(AF_INET6, str, &ipv6_addr);
    if (r == 1)
    {
        big.u = be64toh(*upper);
        big.l = be64toh(*lower);
    }
    else
    {
        triton_uint128_setmax(big);
    }

    return big;
}

/* assign a 64 bit value to a uint128_t */
static inline uint128_t triton_uint128_from_uint64(uint64_t small)
{
    uint128_t big;
    big.u = 0;
    big.l = small;

    return(big);
}

/* increment a 128 bit id by 1 */
static inline uint128_t triton_uint128_inc(uint128_t big)
{
    big.l++;
    if(big.l == 0)
        big.u++;

    return(big);
}

/* returns -1, 0, or 1 if a is less than, equal to, or greater than b */
static inline int triton_cmp_uint128(uint128_t a, uint128_t b)
{
    if(a.u > b.u)
        return 1;
    if(a.u < b.u)
        return -1;

    if(a.l > b.l)
        return 1;
    if(a.l < b.l)
        return -1;

    return 0;
}

#endif /* __TRITON_UINT128_H__ */

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 *
 * vim: ft=c ts=8 sts=4 sw=4 expandtab
 */
