/*
 * (C) 2009 The University of Chicago
 *
 * See COPYRIGHT in top-level directory.
 */

#include <stdlib.h>

#include "triton-list.h"
#include "triton-hash.h"
#include "triton-uint128.h"
#include "test-tap.h"

#include <string.h>

#define DEFAULT_TEST_SIZE 1000

struct e
{
    int v;
    triton_list_link_t link;
};

struct s
{
    char *v;
    triton_list_link_t link;
};

struct n
{
    uint128_t *node;
    triton_list_link_t link;
};

struct k
{
    uint128_t from;
    uint128_t to;
    int32_t tag;
};

struct kv
{
    struct k k;
    struct triton_hash_link link;
};

static int k_compare(const void *k, struct triton_hash_link *l)
{
    const struct k *key = (const struct k *)k;
    struct kv *val = triton_hash_get_entry(l, struct kv, link);
    return (!memcmp(&key->from, &val->k.from, sizeof(key->from)) && !memcmp(&key->to, &val->k.to, sizeof(key->to)) && key->tag == val->k.tag);
}

static int k_hash(const void *k, int table_size)
{
    uint32_t h1 = 0, h2 = 0;
    const struct k *key = (const struct k *)k;
    bj_hashlittle2(key, sizeof(*key), &h1, &h2);
    return h1 & ( table_size - 1);
}

static int e_compare(const void *key, struct triton_hash_link *link)
{
  const int *value1 = (const int *)key;
  struct e *value2 = triton_hash_get_entry(link, struct e, link);
  assert(key);
  assert(link);

  return (value2->v == *value1);
}

static int s_compare(const void *key, struct triton_hash_link *link)
{
    const char *value1 = (const char *)key;
    struct s *value2 = triton_hash_get_entry(link, struct s, link);
    assert(key);
    assert(link);

    return (!strcmp(value2->v, value1));
}

static int n_compare(const void *key, struct triton_hash_link *link)
{
    const uint128_t *value1 = (const uint128_t *)key;
    struct n *value2 = triton_hash_get_entry(link, struct n, link);
    assert(key);
    assert(link);

    return (value2->node->l == value1->l &&
            value2->node->u == value1->u);
}

static void e_destroy(struct e *entry)
{}

static void s_destroy(struct s *entry)
{ free (entry->v); }

int main(int argc, char *argv[])
{
    struct triton_hash_table *table;
    struct e *entry;
    struct e *es;
    struct s *ss;
    struct n *ns;
    struct triton_hash_link *result;
    int i, j, testsize, hash;
    struct triton_hash_table *ktable;
    int testcount, failcount;

    if(argc == 1)
        testsize = DEFAULT_TEST_SIZE;
    else if(argc == 2)
        testsize = atoi(argv[1]);
    else
    {
        fprintf(stderr, "Error: invalid arguments.\n");
        return(-1);
    }

    assert(testsize > 100);

    tap_begin(testcount, failcount);

    es = malloc(sizeof(*es) * testsize);

    /* Begin integer tests... */

    table = triton_hash_init(e_compare, triton_hash_32bit_hash, 16);

    for(i = 0; i < testsize; ++i)
    {
        es[i].v = i;
        triton_list_link_clear(&es[i].link);
    }

    for(i = 0; i < testsize; ++i)
    {
        triton_hash_add(table, &es[i].v, &es[i].link);
    }

    for(i = 0; i < testsize; ++i)
    {
        result = triton_hash_search(table, &i);
        entry = triton_list_get_entry(result, struct e, link);
        if(entry->v != i)
            break;
    }
    tap("triton_hash_search (integers)", i==testsize, testcount, failcount);

    triton_hash_destroy_and_finalize(table, struct e, link, e_destroy);

    /* Begin string tests... */

    for (i = 0; i < 10; i++)
    {
        char buf[512];
        memset (&buf, 0, sizeof(buf));
        snprintf (&buf[0], sizeof (buf)-1, "string: _%d_", i);
        hash = triton_hash_string_hash(&buf[0], 23);
        /* printf("hash(\"%s\") -> %i\n", string->string, hash); */
    }

    table = triton_hash_init(s_compare, triton_hash_string_hash, 23);

    for(i = 0; i < testsize; ++i)
    {
        char buf[512];
        memset (&buf, 0, sizeof(buf));
        snprintf (&buf[0], sizeof(buf)-1, "[%i]", i);
        ss = malloc(sizeof(*ss));
        ss->v = strdup (buf);
        triton_list_link_clear(&ss->link);
        triton_hash_add(table, &buf[0], &(ss->link));
    }

    for(i = 0; i < testsize; ++i)
    {
        char buf[512];
        memset (&buf, 0, sizeof(buf));
        snprintf (&buf[0], sizeof(buf)-1, "[%i]", i);
        result = triton_hash_search(table, &buf[0]);
        assert(result != NULL);
        ss = triton_hash_get_entry(result, struct s, link);
        sscanf(ss->v, "[%i]", &j);
        if(i != j)
            break;
    }
    tap("triton_hash_search (strings)", i==testsize, testcount, failcount);

    triton_hash_destroy_and_finalize(table, struct s, link, s_destroy);

    table = triton_hash_init(n_compare, triton_hash_128bit_hash, 23);

    for(i = 0; i < 3; ++i)
    {
        ns = malloc(sizeof(*ns));
        ns->node = malloc(sizeof(uint128_t));
        ns->node->l = i*i;
        ns->node->u = i+i;
        triton_list_link_clear(&ns->link);
        triton_hash_add(table, ns->node, &(ns->link));
    }

    for(i = 0; i < 3; ++i)
    {
        uint128_t node;
        node.l = i*i;
        node.u = i+i;
        result = triton_hash_search(table, &node);
        assert(result != NULL);
        ns = triton_hash_get_entry(result, struct n, link);
        assert(ns != NULL);
        if(ns->node->l != node.l || ns->node->u != node.u)
            break;
    }
    tap("triton_hash_search (128 bit)", i==3, testcount, failcount);

    ktable = triton_hash_init(k_compare, k_hash, 1024);
    assert(ktable);

    struct kv vals[10];
    for(i = 0; i < 10; ++i)
    {
        triton_list_link_clear(&vals[i].link);
        triton_hash_add(ktable, &vals[i].k, &vals[i].link);
    }

    for(i = 0; i < 10000; ++i)
    {
        assert(triton_hash_search(ktable, &vals[i % 10].k));
    }
    tap("triton_hash_search (opaque key)", 1, testcount, failcount);

    free(es);

    tap_end(testcount, failcount);
    if(failcount)
        return(-1);
    else
        return(0);
}

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 *
 * vim: ts=8 sts=4 sw=4 expandtab
 */
