/*
 * (C) 2009 The University of Chicago
 *
 * See COPYRIGHT in top-level directory.
 */

#include <stdlib.h>
#include <stdio.h>
#include "triton-list.h"
#include "test-tap.h"

#define DEFAULT_TEST_SIZE 1000

struct e
{
    int v;
    triton_list_link_t link;
};

int main(int argc, char *argv[])
{
    triton_list_t l;
    struct e *es, *ep, *esc;
    triton_list_link_t *p, *s;
    int i, testsize, delcount;
    int testcount, failcount;

    if(argc == 1)
        testsize = DEFAULT_TEST_SIZE;
    else if(argc == 2)
        testsize = atoi(argv[1]);
    else
    {
        fprintf(stderr, "Error: invalid arguments.\n");
        return(-1);
    }

    assert(testsize > 100);

    tap_begin(testcount, failcount);

    es = malloc(sizeof(*es) * testsize);

    triton_list_init(&l);

    /* test push/pop */
    for(i = 0; i < testsize; ++i)
    {
        es[i].v = i;
        triton_list_link_clear(&es[i].link);
    }

    for(i = 0; i < testsize; ++i)
    {
        triton_list_add_back(&es[i].link, &l);
    }

    /* test triton_list_del */
    delcount = 0;
    for(i = 0; i < 100; ++i)
    {
        int j = rand() % testsize;

        if(es[j].link.inlist)
        {
            triton_list_del(&es[j].link);
            ++delcount;
        }
    }

    tap("triton_list_del", triton_list_count(&l) == (testsize-delcount),
        testcount, failcount);

    /* test foreach */
    triton_list_for_each(p, s, &l)
    {
        triton_list_del(p);
    }

    tap("triton_list_for_each with triton_list_del", triton_list_empty(&l),
        testcount, failcount);

    /* test push/pop */
    for(i = 0; i < testsize; ++i)
    {
        triton_stack_push(&es[i].link, &l);
    }

    for(i = testsize - 1; i >= 0; --i)
    {
        p = triton_stack_pop(&l);
        if(triton_list_get_entry(p, struct e, link)->v != i)
            break;
    }
    tap("triton_stack_pop", i==-1, testcount, failcount);

    /* test enqueue/dequeue */
    for(i = 0; i < testsize; ++i)
    {
        triton_queue_enqueue(&es[i].link, &l);
    }

    tap("triton_queue_enqueue", triton_list_count(&l) == testsize, testcount, failcount);

    for(i = 0; i < testsize; ++i)
    {
        p = triton_queue_dequeue(&l);
        if(triton_list_get_entry(p, struct e, link)->v != i)
            break;
    }

    tap("triton_queue_dequeue", i == testsize && triton_list_empty(&l), testcount, failcount);

    /* test insert_before */
    for(i = 0; i < 100; ++i)
    {
        triton_queue_enqueue(&es[i].link, &l);
    }

    for(; i < testsize; ++i)
    {
        triton_list_insert_before(&es[i].link, &es[0].link, &l);
    }

    for(i = 0; i < (testsize - 100); ++i)
    {
        assert(triton_list_get_entry(triton_queue_dequeue(&l), struct e, link)->v == 100 + i);
    }

    for(i = 0; i < 100; ++i)
    {
        assert(triton_list_get_entry(triton_queue_dequeue(&l), struct e, link)->v == i);
    }
    /* NOTE: asserts are checking this test case in greater detail above */
    tap("triton_list_insert_before", triton_list_empty(&l), testcount, failcount);

    /* test insert_after */
    for(i = 0; i < 100; ++i)
    {
        triton_queue_enqueue(&es[i].link, &l);
    }

    for(; i < testsize; ++i)
    {
        triton_list_insert_after(&es[i].link, &es[0].link, &l);
    }

    tap("triton_list_insert_after", triton_list_get_entry(triton_queue_dequeue(&l), struct e, link)->v == 0, testcount, failcount);

    for(i = 0; i < (testsize - 100); ++i)
    {
        assert(triton_list_get_entry(triton_queue_dequeue(&l), struct e, link)->v == testsize - 1 - i);
    }

    for(i = 1; i < 100; ++i)
    {
        assert(triton_list_get_entry(triton_queue_dequeue(&l), struct e, link)->v == i);
    }

    /* NOTE: asserts are checking this test case in greater detail above */
    tap("triton_list_insert_after 2", triton_list_empty(&l), testcount, failcount);

    for(i = 0; i < testsize; ++i)
    {
        triton_queue_enqueue(&es[i].link, &l);
    }

    i = 0;
    triton_list_for_each_entry(ep, esc, &l, struct e, link)
    {
        assert(ep->v == i);
        triton_list_del(&ep->link);
        assert(ep->v == i);
        ++i;
    }
    /* NOTE: asserts are checking this test case in greater detail above */
    tap("triton_list_for_each_entry", triton_list_empty(&l), testcount, failcount);

    free(es);

    tap_end(testcount, failcount);
    if(failcount)
        return(-1);
    else
        return(0);
}

/*
 * Local variables:
 *  c-indent-level: 4
 *  c-basic-offset: 4
 * End:
 *
 * vim: ft=c ts=8 sts=4 sw=4 expandtab
 */
